Type: Package
Package: wrswoR
Title: Weighted Random Sampling without Replacement
Version: 1.2.1
Date: 2026-05-13
Authors@R: 
    person(given = "Kirill",
           family = "M\u00fcller",
           role = c("aut", "cre"),
           email = "kirill@cynkra.com",
           comment = c(ORCID = "0000-0002-1416-3412"))
Description: A collection of implementations of classical and
    novel algorithms for weighted sampling without replacement.
License: GPL-3
URL: https://krlmlr.github.io/wrswoR/
BugReports: https://github.com/krlmlr/wrswoR/issues
Depends: R (>= 3.0.2)
Imports: Rcpp
Suggests: BatchExperiments, BiocManager, dplyr, ggplot2, import, kimisc
        (>= 0.2-4), knitcitations, knitr, metap, microbenchmark, otel,
        rmarkdown, roxygen2, rticles (>= 0.1), sampling, testthat (>=
        3.0.0), tidyr, tikzDevice (>= 0.9-1)
LinkingTo: Rcpp (>= 0.11.5)
Encoding: UTF-8
URLNote: https://github.com/krlmlr/wrswoR
Config/gha/extra-packages: metap=?ignore-before-r=4.3.0
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0.9000
NeedsCompilation: yes
Packaged: 2026-05-13 03:42:16 UTC; kirill
Author: Kirill Müller [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1416-3412>)
Maintainer: Kirill Müller <kirill@cynkra.com>
Repository: CRAN
Date/Publication: 2026-05-13 05:10:09 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-13 05:22:22 UTC; unix
Archs: wrswoR.so.dSYM

# Generated by roxygen2: do not edit by hand

export(sample_int_R)
export(sample_int_cccrank)
export(sample_int_ccrank)
export(sample_int_crank)
export(sample_int_expj)
export(sample_int_expjs)
export(sample_int_rank)
export(sample_int_rej)
importFrom(Rcpp,evalCpp)
importFrom(stats,rexp)
importFrom(utils,head)
useDynLib(wrswoR, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# wrswoR

<!-- badges: start -->

[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![R build
status](https://github.com/krlmlr/wrswoR/workflows/rcc/badge.svg)](https://github.com/krlmlr/wrswoR/actions)
[![Codecov test
coverage](https://codecov.io/gh/krlmlr/wrswoR/branch/main/graph/badge.svg)](https://app.codecov.io/gh/krlmlr/wrswoR?branch=main)
[![CRAN
status](https://www.r-pkg.org/badges/version/wrswoR)](https://CRAN.R-project.org/package=wrswoR)
<!-- badges: end -->

The goal of wrswoR is to provide faster implementations of weighted
random sampling without replacement in R.

## Installation

You can install the released version of wrswoR from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("wrswoR")
```

And the development version from [GitHub](https://github.com/) with:

``` r
# install.packages("devtools")
devtools::install_github("krlmlr/wrswoR")
```

## Example

The functions in this package are a drop-in replacement to
`sample.int(n, size, replace = FALSE, prob = prob)`. With large `n`,
`sample.int()` becomes too slow to be practical, unlike the functions in
this package.

``` r
library(wrswoR)
set.seed(20200726)

sample_int_crank(20, 10, 1:20)
#>  [1]  8 18 14 17 11 15 10  4 13  5
```

#' @rdname sample_int
#' @details `sample_int_rank()`, `sample_int_crank()`,
#'   `sample_int_ccrank()` and `sample_int_cccrank()` implement one-pass random
#'   sampling (Efraimidis and Spirakis, 2006, Algorithm A; see also Yellott,
#'   1977, and Vieira, 2014, for an equivalent formulation).  The first function
#'   is implemented purely in R, the other three are optimized `Rcpp`
#'   implementations (`*_crank()` uses R vectors internally, while
#'   `*_ccrank()` uses `std::vector`; `*_cccrank()` is a memory-optimized
#'   implementation that only requires \eqn{O(\text{size})} extra space;
#'   surprisingly, `*_crank()` seems to be faster on most inputs). It can be
#'   shown that the order statistic of \eqn{U^{(1/w_i)}} has the same
#'   distribution as random sampling without replacement (\eqn{U=\mbox{uniform}(0,1)}{U=uniform(0,1)}
#'   distribution). To increase numerical stability, \eqn{\log(U) /
#'   w_i}{log(U) / w_i} is computed instead; the log transform does not
#'   change the order statistic.
#' @author Dinre (for `*_rank()`), Kirill Müller
#'   (for all other functions)
#' @references Efraimidis, Pavlos S., and Paul G. Spirakis. "Weighted
#' random sampling with a reservoir." *Information Processing
#' Letters* 97, no. 5 (2006): 181-185.
#'
#' John I. Yellott. "The relationship between Luce’s choice axiom,
#' Thurstone’s theory of comparative judgment, and the double exponential
#' distribution." *Journal of Mathematical Psychology*, 15(2):109 – 144, 1977.
#'
#' Vieira, T. Gumbel-max trick and weighted reservoir sampling, 2014. URL
#' <https://timvieira.github.io/blog/post/2014/08/01/gumbel-max-trick-and-weighted-reservoir-sampling/>.
#'
#' @export
#' @examples
#' ## Algorithm A
#' s <- sample_int_rank(20000, 10000, runif(20000))
#' stopifnot(unique(s) == s)
#' p <- c(995, rep(1, 5))
#' n <- 1000
#' set.seed(42)
#' tbl <- table(replicate(sample_int_rank(6, 3, p),
#'                        n = n)) / n
#' stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)
#'
#' @importFrom stats rexp
#' @importFrom utils head
sample_int_rank <- function(n, size, prob) {
  .check_args(n, size, prob)
  head(order(rexp(n) / prob), size)
}

.check_args <- function(n, size, prob) {
  if (is.na(n)) {
    stop("'n' must not be NA", call. = FALSE)
  }
  
  if (is.na(size)) {
    stop("'size' must not be NA", call. = FALSE)
  }
  
  if (n < 0) {
    stop("'n' must be non-negative", call. = FALSE)
  }
  
  if (size < 0) {
    stop("'size' must be non-negative", call. = FALSE)
  }
  
  if (n < size) {
    stop("cannot take a sample larger than the population", call. = FALSE)
  }

  if (length(prob) != n) {
    stop("incorrect number of probabilities", call. = FALSE)
  }
}

#' @rdname sample_int
#'
#' @details `sample_int_rej()` uses repeated weighted sampling with
#'   replacement and a variant of rejection sampling. It is implemented purely
#'   in R.
#'   This function simulates weighted sampling without replacement using
#'   somewhat more draws *with* replacement, and then discarding
#'   duplicate values (rejection sampling).  If too few items are
#'   sampled, the routine calls itself recursively on a (hopefully) much
#'   smaller problem.  See also
#'   <https://stats.stackexchange.com/q/20590/6432>.
#' @export
#' @examples
#' ## Rejection sampling
#' s <- sample_int_rej(20000, 10000, runif(20000))
#' stopifnot(unique(s) == s)
#' p <- c(995, rep(1, 5))
#' n <- 1000
#' set.seed(42)
#' tbl <- table(replicate(sample_int_rej(6, 3, p),
#'                        n = n)) / n
#' stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)
#'
sample_int_rej <- function(n, size, prob) {
  .check_args(n, size, prob)
  .sample_int_rej(n, size, prob, 2, 1)
}

# Euler-Mascheroni constant
.EM = 0.57721566490153286060651209008240243104215933593992

# Computes the harmonic series. Exact for the first
# .harmonic.series.max values (through table lookup), otherwise using
# the approximation ln(a) + \gamma + 1 / (2a). Source:
# https://en.wikipedia.org/wiki/Harmonic_number
.harmonic <- function(a) {
  stopifnot(a >= 0)
  if (a < length(.harmonic.series)) {
    .harmonic.series[a + 1]
  } else {
    log(a) + .EM + .5 / a
  }
}

# Workhorse
.sample_int_rej <- function(
  n, size, prob, MAX_OVERSHOOT, BIAS) {

  span <- start_local_active_span(
    ".sample_int_rej",
    attributes = list(n = n, size = size, prob_length = length(prob))
  )

  # How many draws *with replacement* are required on average, assuming
  # *uniform* weights? (With non-uniform weights, this number can only
  # increase.) The result is a general case of the coupon collector
  # problem, see https://math.stackexchange.com/q/247569/16420 for an
  # analysis. BIAS can be supplied to correct the estimate by a factor,
  # at most n * MAX_OVERSHOOT samples will be drawn.  Both are tuning
  # parameters, ideal values are still to be found through simulation.
  wr.size <- ceiling(n * min(BIAS * (.harmonic(n) - .harmonic(n - size)),
                             MAX_OVERSHOOT))
  span$set_attribute("wr.size", wr.size)

  # Do the sampling with replacement...
  wr.sample <- sample.int(n, size=wr.size, replace=T, prob)
  # ...but keep only unique values.
  wr.sample <- unique(wr.sample)
  wr.sample.len <- length(wr.sample)
  span$set_attribute("wr.sample.len", wr.sample.len)

  # How much still left to do?
  rem.size <- size - wr.sample.len
  # Done? Great!
  if (rem.size <= 0)
    return (head(wr.sample, size))

  # Not yet: Find out which indexes haven't been sampled yet.  Recall
  # that negative indexes in a vector subscription mean "all but
  # the selected".
  rem.indexes <- (1:n)[-wr.sample]
  rem.n <- length(rem.indexes)
  stopifnot(rem.n == n - wr.sample.len)

  # Recursive call to sample without replacement from the remaining
  # weights
  rem.sample <- .sample_int_rej(rem.n, rem.size,
                                prob[rem.indexes],
                                MAX_OVERSHOOT, BIAS)

  # Combine the results, substitute the indexes from 1:rem.n obtained
  # from the recursive call using the rem.indexes map
  c(wr.sample, rem.indexes[rem.sample])
}

test_that("sample_int_rej emits an otel span with parameter and result attributes", {
  skip_if_not_installed("otel")

  recorder <- make_span_recorder()
  local_mocked_bindings(start_local_active_span = recorder$start)

  set.seed(42)
  sample_int_rej(6, 3, c(995, rep(1, 5)))

  expect_snapshot(recorder$spans)
})

test_that("sample_int_rej recurses and emits one span per recursion", {
  skip_if_not_installed("otel")

  recorder <- make_span_recorder()
  local_mocked_bindings(start_local_active_span = recorder$start)

  set.seed(20150618L)
  sample_int_rej(50, 40, runif(50))

  expect_snapshot(lapply(recorder$spans, function(s) {
    list(name = s$name, attribute_keys = sort(names(s$attributes)))
  }))
})

test_that("otel_tracer_name follows the reverse-URL convention", {
  expect_equal(
    asNamespace("wrswoR")$otel_tracer_name,
    "io.github.krlmlr.wrswoR"
  )
})

#' @name sample_int_expjs
#' @rdname sample_int
#' @examples
#' ## Algorithm A-ExpJ (paper version)
#' s <- sample_int_expjs(20000, 10000, runif(20000))
#' stopifnot(unique(s) == s)
#' p <- c(995, rep(1, 5))
#' n <- 1000
#' set.seed(42)
#' tbl <- table(replicate(sample_int_expjs(6, 3, p),
#'                        n = n)) / n
#' stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)
#'
NULL

#' Weighted sampling without replacement
#'
#' These functions implement weighted sampling without replacement using various
#' algorithms, i.e., they take a sample of the specified
#' `size` from the elements of `1:n` without replacement, using the
#' weights defined by `prob`.  The call
#' `sample_int_*(n, size, prob)` is equivalent
#' to `sample.int(n, size, replace = F, prob)`.  (The results will
#' most probably be different for the same random seed, but the
#' returned samples are distributed identically for both calls.)
#' Except for `sample_int_R()` (which
#' has quadratic complexity as of this writing), all functions have complexity
#' \eqn{O(n \log n)}{O(n log n)} or better and
#' often run faster than R's implementation, especially when `n` and
#' `size` are large.
#'
#' @details
#'   `sample_int_R()` is a simple wrapper for [base::sample.int()].
#'
#' @inheritParams base::sample.int
#' @return An integer vector of length `size` with elements from
#'   `1:n`.
#' @seealso [base::sample.int()]
#' @references <https://stackoverflow.com/q/15113650/946850>
#' @name sample_int
#' @export
#' @examples
#' # Base R implementation
#' s <- sample_int_R(2000, 1000, runif(2000))
#' stopifnot(unique(s) == s)
#' p <- c(995, rep(1, 5))
#' n <- 1000
#' set.seed(42)
#' tbl <- table(replicate(sample_int_R(6, 3, p),
#'                        n = n)) / n
#' stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)
#'
sample_int_R <- function(n, size, prob) {
  sample.int(n, size, replace = FALSE, prob)
}

otel_tracer_name <- "io.github.krlmlr.wrswoR"

# Calls otel::start_local_active_span() when otel is installed, otherwise
# returns a span object whose methods are no-ops. parent.frame() pins the
# activation scope to the caller of this wrapper so the span ends with the
# caller, not with this function.
start_local_active_span <- function(name, attributes = NULL, ...) {
  if (requireNamespace("otel", quietly = TRUE)) {
    otel::start_local_active_span(
      name,
      attributes = attributes,
      ...,
      activation_scope = parent.frame()
    )
  } else {
    list(set_attribute = function(key, value) invisible(NULL))
  }
}

#' Record otel spans emitted by wrswoR functions.
#'
#' Internal test helper. Use together with
#' [testthat::local_mocked_bindings()] to swap `start_local_active_span`
#' for the recorder's capture function.
#'
#' @noRd
make_span_recorder <- function() {
  recorder <- new.env(parent = emptyenv())
  recorder$spans <- list()

  recorder$start <- function(name, attributes = NULL, ...) {
    idx <- length(recorder$spans) + 1L
    recorder$spans[[idx]] <- list(
      name = name,
      attributes = as.list(attributes)
    )
    list(
      set_attribute = function(key, value) {
        recorder$spans[[idx]]$attributes[[key]] <- value
      }
    )
  }

  recorder
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' @rdname sample_int
#' @export
sample_int_crank <- function(n, size, prob) {
    .Call(`_wrswoR_sample_int_crank`, n, size, prob)
}

#' @rdname sample_int
#' @export
sample_int_ccrank <- function(n, size, prob) {
    .Call(`_wrswoR_sample_int_ccrank`, n, size, prob)
}

#' @rdname sample_int
#' @export
sample_int_cccrank <- function(n, size, prob) {
    .Call(`_wrswoR_sample_int_cccrank`, n, size, prob)
}

#' @rdname sample_int
#' @export
sample_int_expj <- function(n, size, prob) {
    .Call(`_wrswoR_sample_int_expj`, n, size, prob)
}

#' @rdname sample_int
#' @export
sample_int_expjs <- function(n, size, prob) {
    .Call(`_wrswoR_sample_int_expjs`, n, size, prob)
}

#' @name sample_int_expj
#' @rdname sample_int
#' @details `sample_int_expj()` and `sample_int_expjs()`
#'   implement one-pass random sampling with a reservoir with exponential jumps
#'   (Efraimidis and Spirakis, 2006, Algorithm A-ExpJ).  Both functions are
#'   implemented in `Rcpp`; `*_expj()` uses log-transformed keys,
#'   `*_expjs()` implements the algorithm in the paper verbatim
#'   (at the cost of numerical stability).
#' @examples
#' ## Algorithm A-ExpJ (with log-transformed keys)
#' s <- sample_int_expj(20000, 10000, runif(20000))
#' stopifnot(unique(s) == s)
#' p <- c(995, rep(1, 5))
#' n <- 1000
#' set.seed(42)
#' tbl <- table(replicate(sample_int_expj(6, 3, p),
#'                        n = n)) / n
#' stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)
#'
NULL

#' @useDynLib wrswoR, .registration = TRUE

#' @title Faster weighted sampling without replacement
#' @description \R's default sampling without replacement using
#'  [base::sample.int()] seems to require quadratic run time,
#'  e.g., when using weights drawn from a uniform distribution. For large
#'  sample sizes, this is too slow.  This package contains several
#'  alternative implementations.
#' @details Implementations are adapted from
#'  <https://stackoverflow.com/q/15113650/946850>.
#'
#' @name wrswoR-package
#' @aliases wrswoR-package wrswoR
#' @docType package
#' @author Kirill Müller
#' @references Efraimidis, Pavlos S., and Paul G. Spirakis. "Weighted
#' random sampling with a reservoir." *Information Processing Letters* 97,
#' no. 5 (2006): 181-185.
#'
#' Wong, Chak-Kuen, and Malcolm C. Easton. "An efficient method for
#' weighted sampling without replacement." *SIAM Journal on Computing* 9,
#' no. 1 (1980): 111-113.
#'
#'
#' @keywords package
#' @examples
#' sample_int_rej(100, 50, 1:100)
NULL

test_that("no dupes", {
  for (funcname in funcnames) {
    for (i in 1:20) {
      expect_equal(sum(duplicated(funcs[[funcname]](i, i - 1, seq_len(i)))), 0,
                  label = funcname)
    }
  }
})

test_that("all between 1 and n", {
  for (funcname in funcnames) {
    for (i in 1:20) {
      expect_true(all(funcs[[funcname]](i, i - 1, seq_len(i)) %in% seq_len(i)),
                  label = funcname)
    }
  }
})

test_that("sample larger than population", {
  for (funcname in funcnames) {
    expect_error(funcs[[funcname]](1, 2, 1), "larger than the population", label = funcname)
  }
})

# This will fail if Sys.getlocale(category = "LC_MESSAGES") is not "en_*.*"
test_that("number of probabilities", {
  for (funcname in funcnames) {
    expect_error(funcs[[funcname]](3, 2, 1:2), "incorrect number of probabilities", label = funcname)
    expect_error(funcs[[funcname]](3, 2, 1:4), "incorrect number of probabilities", label = funcname)
    expect_error(funcs[[funcname]](4, 3, 1:5), "incorrect number of probabilities", label = funcname)
  }
})

test_that("NA sample size", {
  for (funcname in funcnames) {
    # sample_int_R uses base::sample.int which has different error messages
    if (funcname == "sample_int_R") {
      expect_error(funcs[[funcname]](5, NA, rep(1, 5)), "invalid.*size", label = funcname)
    } else {
      expect_error(funcs[[funcname]](5, NA, rep(1, 5)), "'size' must not be NA", label = funcname)
    }
  }
})

test_that("NA population size", {
  for (funcname in funcnames) {
    # sample_int_R uses base::sample.int which has different error messages
    if (funcname == "sample_int_R") {
      expect_error(funcs[[funcname]](NA, 3, rep(1, 5)), "invalid.*first", label = funcname)
    } else {
      expect_error(funcs[[funcname]](NA, 3, rep(1, 5)), "'n' must not be NA", label = funcname)
    }
  }
})

test_that("negative sample size", {
  for (funcname in funcnames) {
    # sample_int_R uses base::sample.int which has different error messages
    if (funcname == "sample_int_R") {
      expect_error(funcs[[funcname]](5, -1, rep(1, 5)), "invalid.*size", label = funcname)
    } else {
      expect_error(funcs[[funcname]](5, -1, rep(1, 5)), "'size' must be non-negative", label = funcname)
    }
  }
})

test_that("negative population size", {
  for (funcname in funcnames) {
    # sample_int_R uses base::sample.int which has different error messages
    if (funcname == "sample_int_R") {
      expect_error(funcs[[funcname]](-1, 0, rep(1, 5)), "invalid.*first", label = funcname)
    } else {
      expect_error(funcs[[funcname]](-1, 0, rep(1, 5)), "'n' must be non-negative", label = funcname)
    }
  }
})

test_that("zero-length output", {
  for (funcname in funcnames) {
    expect_equal(funcs[[funcname]](10, 0, 1:10), integer(), label = funcname)
  }
})

test_that("uniform probability", {
  n <- 13
  prob <- rep(1, n)
  for (funcname in funcnames) {
    expect_equal(sort(funcs[[funcname]](n, n, prob)), seq_len(n), label = funcname)
  }
})

test_that("heavily skewed probability", {
  prob <- c(rep(1, 10), 1e10)
  for (funcname in funcnames) {
    expect_equal(funcs[[funcname]](length(prob), 1, prob), length(prob), label = funcname)
  }
})

test_that("heavily skewed probability (reversed)", {
  prob <- c(1e10, rep(1, 10))
  for (funcname in funcnames) {
    expect_equal(funcs[[funcname]](length(prob), 1, prob), 1, label = funcname)
  }
})

# sample_int_rej emits an otel span with parameter and result attributes

    Code
      recorder$spans
    Output
      [[1]]
      [[1]]$name
      [1] ".sample_int_rej"
      
      [[1]]$attributes
      [[1]]$attributes$n
      [1] 6
      
      [[1]]$attributes$size
      [1] 3
      
      [[1]]$attributes$prob_length
      [1] 6
      
      [[1]]$attributes$wr.size
      [1] 4
      
      [[1]]$attributes$wr.sample.len
      [1] 1
      
      
      
      [[2]]
      [[2]]$name
      [1] ".sample_int_rej"
      
      [[2]]$attributes
      [[2]]$attributes$n
      [1] 5
      
      [[2]]$attributes$size
      [1] 2
      
      [[2]]$attributes$prob_length
      [1] 5
      
      [[2]]$attributes$wr.size
      [1] 3
      
      [[2]]$attributes$wr.sample.len
      [1] 2
      
      
      

# sample_int_rej recurses and emits one span per recursion

    Code
      lapply(recorder$spans, function(s) {
        list(name = s$name, attribute_keys = sort(names(s$attributes)))
      })
    Output
      [[1]]
      [[1]]$name
      [1] ".sample_int_rej"
      
      [[1]]$attribute_keys
      [1] "n"             "prob_length"   "size"          "wr.sample.len"
      [5] "wr.size"      
      
      
      [[2]]
      [[2]]$name
      [1] ".sample_int_rej"
      
      [[2]]$attribute_keys
      [1] "n"             "prob_length"   "size"          "wr.sample.len"
      [5] "wr.size"      
      
      
      [[3]]
      [[3]]$name
      [1] ".sample_int_rej"
      
      [[3]]$attribute_keys
      [1] "n"             "prob_length"   "size"          "wr.sample.len"
      [5] "wr.size"      
      
      
      [[4]]
      [[4]]$name
      [1] ".sample_int_rej"
      
      [[4]]$attribute_keys
      [1] "n"             "prob_length"   "size"          "wr.sample.len"
      [5] "wr.size"      
      
      

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrswoR-package.R
\docType{package}
\name{wrswoR-package}
\alias{wrswoR-package}
\alias{wrswoR}
\title{Faster weighted sampling without replacement}
\description{
\R's default sampling without replacement using
\code{\link[base:sample.int]{base::sample.int()}} seems to require quadratic run time,
e.g., when using weights drawn from a uniform distribution. For large
sample sizes, this is too slow.  This package contains several
alternative implementations.
}
\details{
Implementations are adapted from
\url{https://stackoverflow.com/q/15113650/946850}.
}
\examples{
sample_int_rej(100, 50, 1:100)
}
\references{
Efraimidis, Pavlos S., and Paul G. Spirakis. "Weighted
random sampling with a reservoir." \emph{Information Processing Letters} 97,
no. 5 (2006): 181-185.

Wong, Chak-Kuen, and Malcolm C. Easton. "An efficient method for
weighted sampling without replacement." \emph{SIAM Journal on Computing} 9,
no. 1 (1980): 111-113.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://krlmlr.github.io/wrswoR/}
  \item Report bugs at \url{https://github.com/krlmlr/wrswoR/issues}
}

}
\author{
Kirill Müller
}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/sample_int_R.R,
%   R/sample_int_ccrank.R, R/sample_int_crank.R, R/sample_int_expj.R,
%   R/sample_int_expjs.R, R/sample_int_rank.R, R/sample_int_rej.R
\name{sample_int_crank}
\alias{sample_int_crank}
\alias{sample_int_ccrank}
\alias{sample_int_cccrank}
\alias{sample_int_expj}
\alias{sample_int_expjs}
\alias{sample_int}
\alias{sample_int_R}
\alias{sample_int_rank}
\alias{sample_int_rej}
\title{Weighted sampling without replacement}
\usage{
sample_int_crank(n, size, prob)

sample_int_ccrank(n, size, prob)

sample_int_cccrank(n, size, prob)

sample_int_expj(n, size, prob)

sample_int_expjs(n, size, prob)

sample_int_R(n, size, prob)

sample_int_rank(n, size, prob)

sample_int_rej(n, size, prob)
}
\arguments{
\item{n}{a positive number, the number of items to choose from.  See
    \sQuote{Details.}}

\item{size}{a non-negative integer giving the number of items to choose.}

\item{prob}{a vector of probability weights for obtaining the elements
    of the vector being sampled.}
}
\value{
An integer vector of length \code{size} with elements from
\code{1:n}.
}
\description{
These functions implement weighted sampling without replacement using various
algorithms, i.e., they take a sample of the specified
\code{size} from the elements of \code{1:n} without replacement, using the
weights defined by \code{prob}.  The call
\verb{sample_int_*(n, size, prob)} is equivalent
to \code{sample.int(n, size, replace = F, prob)}.  (The results will
most probably be different for the same random seed, but the
returned samples are distributed identically for both calls.)
Except for \code{sample_int_R()} (which
has quadratic complexity as of this writing), all functions have complexity
\eqn{O(n \log n)}{O(n log n)} or better and
often run faster than R's implementation, especially when \code{n} and
\code{size} are large.
}
\details{
\code{sample_int_R()} is a simple wrapper for \code{\link[base:sample.int]{base::sample.int()}}.

\code{sample_int_expj()} and \code{sample_int_expjs()}
implement one-pass random sampling with a reservoir with exponential jumps
(Efraimidis and Spirakis, 2006, Algorithm A-ExpJ).  Both functions are
implemented in \code{Rcpp}; \verb{*_expj()} uses log-transformed keys,
\verb{*_expjs()} implements the algorithm in the paper verbatim
(at the cost of numerical stability).

\code{sample_int_rank()}, \code{sample_int_crank()},
\code{sample_int_ccrank()} and \code{sample_int_cccrank()} implement one-pass random
sampling (Efraimidis and Spirakis, 2006, Algorithm A; see also Yellott,
1977, and Vieira, 2014, for an equivalent formulation).  The first function
is implemented purely in R, the other three are optimized \code{Rcpp}
implementations (\verb{*_crank()} uses R vectors internally, while
\verb{*_ccrank()} uses \code{std::vector}; \verb{*_cccrank()} is a memory-optimized
implementation that only requires \eqn{O(\text{size})} extra space;
surprisingly, \verb{*_crank()} seems to be faster on most inputs). It can be
shown that the order statistic of \eqn{U^{(1/w_i)}} has the same
distribution as random sampling without replacement (\eqn{U=\mbox{uniform}(0,1)}{U=uniform(0,1)}
distribution). To increase numerical stability, \eqn{\log(U) /
  w_i}{log(U) / w_i} is computed instead; the log transform does not
change the order statistic.

\code{sample_int_rej()} uses repeated weighted sampling with
replacement and a variant of rejection sampling. It is implemented purely
in R.
This function simulates weighted sampling without replacement using
somewhat more draws \emph{with} replacement, and then discarding
duplicate values (rejection sampling).  If too few items are
sampled, the routine calls itself recursively on a (hopefully) much
smaller problem.  See also
\url{https://stats.stackexchange.com/q/20590/6432}.
}
\examples{
# Base R implementation
s <- sample_int_R(2000, 1000, runif(2000))
stopifnot(unique(s) == s)
p <- c(995, rep(1, 5))
n <- 1000
set.seed(42)
tbl <- table(replicate(sample_int_R(6, 3, p),
                       n = n)) / n
stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)

## Algorithm A, Rcpp version using std::vector
s <- sample_int_ccrank(20000, 10000, runif(20000))
stopifnot(unique(s) == s)
p <- c(995, rep(1, 5))
n <- 1000
set.seed(42)
tbl <- table(replicate(sample_int_ccrank(6, 3, p),
                       n = n)) / n
stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)

## Algorithm A, Rcpp version using R vectors
s <- sample_int_crank(20000, 10000, runif(20000))
stopifnot(unique(s) == s)
p <- c(995, rep(1, 5))
n <- 1000
set.seed(42)
tbl <- table(replicate(sample_int_crank(6, 3, p),
                       n = n)) / n
stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)

## Algorithm A-ExpJ (with log-transformed keys)
s <- sample_int_expj(20000, 10000, runif(20000))
stopifnot(unique(s) == s)
p <- c(995, rep(1, 5))
n <- 1000
set.seed(42)
tbl <- table(replicate(sample_int_expj(6, 3, p),
                       n = n)) / n
stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)

## Algorithm A-ExpJ (paper version)
s <- sample_int_expjs(20000, 10000, runif(20000))
stopifnot(unique(s) == s)
p <- c(995, rep(1, 5))
n <- 1000
set.seed(42)
tbl <- table(replicate(sample_int_expjs(6, 3, p),
                       n = n)) / n
stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)

## Algorithm A
s <- sample_int_rank(20000, 10000, runif(20000))
stopifnot(unique(s) == s)
p <- c(995, rep(1, 5))
n <- 1000
set.seed(42)
tbl <- table(replicate(sample_int_rank(6, 3, p),
                       n = n)) / n
stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)

## Rejection sampling
s <- sample_int_rej(20000, 10000, runif(20000))
stopifnot(unique(s) == s)
p <- c(995, rep(1, 5))
n <- 1000
set.seed(42)
tbl <- table(replicate(sample_int_rej(6, 3, p),
                       n = n)) / n
stopifnot(abs(tbl - c(1, rep(0.4, 5))) < 0.04)

}
\references{
\url{https://stackoverflow.com/q/15113650/946850}

Efraimidis, Pavlos S., and Paul G. Spirakis. "Weighted
random sampling with a reservoir." \emph{Information Processing
Letters} 97, no. 5 (2006): 181-185.

John I. Yellott. "The relationship between Luce’s choice axiom,
Thurstone’s theory of comparative judgment, and the double exponential
distribution." \emph{Journal of Mathematical Psychology}, 15(2):109 – 144, 1977.

Vieira, T. Gumbel-max trick and weighted reservoir sampling, 2014. URL
\url{https://timvieira.github.io/blog/post/2014/08/01/gumbel-max-trick-and-weighted-reservoir-sampling/}.
}
\seealso{
\code{\link[base:sample.int]{base::sample.int()}}
}
\author{
Dinre (for \verb{*_rank()}), Kirill Müller
(for all other functions)
}
